/*
 * Copyright 2010 University of Chicago/Argonne National Laboratory
 *
 * @author Neill Miller
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "soapH.h"
#include "uuid/uuid.h"
#include "SamlAuthzServiceSOAP.nsmap"
#include "libglobus_adq.h"

/* uncomment for a little more verbose output */
/* #define ADQ_DEBUG */

/*
  Input parameters:
  char *service_endpoint: the service endpoint (e.g. "http://www.localhost:8080/service")
  char *openid: the user's OpenID (e.g. "http://my.openid")
  char *resource: the resource we're requesting access to (e.g. "/PATH/TO/FILE")
  char *action: the action we're requesting of the resource (e.g. "read")

  Output parameter:
  authz_decision_t *authz_out_p: the struct containing the parsed
  contents of the authz decision statement return from the server (on
  success)

  Return value: -1 on failure, 0 on success
*/
int get_authorization_decision(
    char *service_endpoint, char *openid, char *resource,
    char *action, authz_decision_t *authz_out_p)
{
    int ret = -1;
    uuid_t uuid;
    struct soap soap;
    char *x509_cert_dir = NULL;
    char buf[MAX_UUID_LEN] = {0};
    struct ns1__AuthzDecisionQueryType authzDecisionQuery;
    struct ns1__ResponseType response;


#ifdef ADQ_DEBUG
    printf("%s %s %s %s\n", service_endpoint, openid, resource, action);
#endif
    /* initialize soap library */
    soap_init(&soap);

    x509_cert_dir = getenv("X509_CERT_DIR");
    soap.capath = (x509_cert_dir ? x509_cert_dir : "/etc/grid-security/certificates");

#ifdef ADQ_DEBUG
    printf("Using CA Path: %s\n", soap.capath);
#endif

    if (service_endpoint && openid && resource && action)
    {
        memset(&authzDecisionQuery, 0, sizeof(authzDecisionQuery));
        memset(&response, 0, sizeof(response));

        if (authz_out_p)
        {
            authz_out_p->decision = ADQ_DENY;
        }

        /* populate the authzDecisionQuery */
        authzDecisionQuery.ns2__Subject = calloc(1, sizeof(struct ns2__SubjectType));
        authzDecisionQuery.ns2__Subject->__union_SubjectType = SOAP_UNION__ns2__union_SubjectType_ns2__NameID;
        authzDecisionQuery.ns2__Subject->union_SubjectType.ns2__NameID = calloc(1, sizeof(struct ns2__NameIDType));
        authzDecisionQuery.ns2__Subject->union_SubjectType.ns2__NameID->__item = openid;
        authzDecisionQuery.ns2__Subject->union_SubjectType.ns2__NameID->Format = strdup("urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified");

        uuid_generate_random(uuid);
        snprintf(buf, MAX_UUID_LEN, "%llu", (long long unsigned int)uuid);
        authzDecisionQuery.ID = strdup(buf);
        authzDecisionQuery.Version = strdup("2.0");
        authzDecisionQuery.IssueInstant = time(NULL);
        authzDecisionQuery.Resource = resource;

        authzDecisionQuery.__sizeAction = 1;
        authzDecisionQuery.ns2__Action = calloc(1, sizeof(struct ns2__ActionType));
        /*
          only Read and Write actions are supported (according to ESG
          ICD); this is enforced by using "Read" for all actions
          passed in except for "Write"
        */
        if (strcasecmp(action, "write") == 0)
        {
            authzDecisionQuery.ns2__Action->__item = "Write";
        }
        else
        {
            authzDecisionQuery.ns2__Action->__item = "Read";
        }
#ifdef ADQ_DEBUG
        printf("Using normalized Action of %s\n", authzDecisionQuery.ns2__Action->__item);
#endif
        authzDecisionQuery.ns2__Action->Namespace = strdup("urn:oasis:names:tc:SAML:2.0:assertion");

#ifdef ADQ_DEBUG
        printf("About to do soap call ...");
#endif
        /* do the soap call */
        if (soap_call___ns1__getAuthzDecision(
                &soap, service_endpoint, NULL, &authzDecisionQuery, &response) != SOAP_OK)
        {
            soap_print_fault(&soap, stderr);
#ifdef ADQ_DEBUG
            printf("done (Error)\n");
#endif
            goto end;
        }
#ifdef ADQ_DEBUG
        printf("done (Success)\n");
#endif

        /* return success if we got a response back; caller will have to check authz_out_p->decision */
        ret = 0;

#ifdef ADQ_DEBUG
        printf("Response Status %s, ID %s, InResponseTo %s, IssueInstant %lu\n",
               response.ns1__Status->ns1__StatusCode->Value, response.ID, response.InResponseTo,
               (response.IssueInstant ? response.IssueInstant : 0));

        if (response.__union_ResponseType)
        {
            printf("Response type %s\n",
                   ((response.__union_ResponseType->__union_ResponseType == SOAP_UNION__ns1__union_ResponseType_ns2__Assertion) ?
                    "Assertion" : "Encrypted Assertion"));
        }
#endif

        if ((response.__union_ResponseType != 0) &&
            (response.__union_ResponseType->__union_ResponseType == SOAP_UNION__ns1__union_ResponseType_ns2__Assertion))
        {
#ifdef ADQ_DEBUG
            printf("Assertion ID %s, Version %s, IssueInstant %lu, AssertionType %s\n",
                   response.__union_ResponseType->union_ResponseType.ns2__Assertion->ID,
                   response.__union_ResponseType->union_ResponseType.ns2__Assertion->Version,
                   response.__union_ResponseType->union_ResponseType.ns2__Assertion->IssueInstant,
                   ((response.__union_ResponseType->union_ResponseType.ns2__Assertion->__union_AssertionType->__union_AssertionType == SOAP_UNION__ns2__union_AssertionType_ns2__AuthzDecisionStatement) ? "AuthzDecisionStatement" : "Other"));
#endif
            struct ns2__AuthzDecisionStatementType *authzDecisionStmt = response.__union_ResponseType->union_ResponseType.ns2__Assertion->__union_AssertionType->union_AssertionType.ns2__AuthzDecisionStatement;

#ifdef ADQ_DEBUG
            printf("AuthzDecisionStatement: Resource %s, Action %s, Decision %s\n", authzDecisionStmt->Resource,
                   authzDecisionStmt->ns2__Action->__item, ((authzDecisionStmt->Decision == ns2__DecisionType__Permit) ? "PERMIT" : "NOT PERMIT"));
#endif

            /* if all went well, populate the outgoing authz_decision parameter */
            if (authz_out_p)
            {
                strncpy(authz_out_p->action, authzDecisionStmt->ns2__Action->__item, MAX_ACTION_LEN);
                if (authzDecisionStmt->Resource)
                {
                    strncpy(authz_out_p->resource, authzDecisionStmt->Resource, MAX_RESOURCE_LEN);
                }
                else
                {
                    strncpy(authz_out_p->resource, resource, MAX_RESOURCE_LEN);
                }
                switch(authzDecisionStmt->Decision)
                {
                    case ns2__DecisionType__Permit:
                        authz_out_p->decision = ADQ_PERMIT;
                        break;
                    case ns2__DecisionType__Deny:
                        authz_out_p->decision = ADQ_DENY;
                        break;
                    default:
                        authz_out_p->decision = ADQ_INDETERMINATE;
                        break;
                }
            }
        }
#ifdef ADQ_DEBUG
        else
        {
            printf("Unexpected Response type received (NOT an Assertion).\n");
        }
#endif
    }

  end:

/*     ADQ_FREE(authzDecisionQuery.ns2__Subject->union_SubjectType.ns2__NameID->Format); */
/*     ADQ_FREE(authzDecisionQuery.ns2__Subject->union_SubjectType.ns2__NameID); */
    ADQ_FREE(authzDecisionQuery.ID);
    ADQ_FREE(authzDecisionQuery.Version);
/*     ADQ_FREE(authzDecisionQuery.ns2__Action->Namespace); */
    ADQ_FREE(authzDecisionQuery.ns2__Action);

    soap_destroy(&soap);
    soap_end(&soap);
    soap_done(&soap);

    return ret;
}

/* this is a hack, but is used to resolve unused/unresolved symbols */
int soap_in_PointerTo_XML() { return 0; }
int soap_out_PointerTo_XML() { return 0; }
int SOAP_FMAC6 __ns1__getAuthzDecision(struct soap* s, struct ns1__AuthzDecisionQueryType *ns1__AuthzDecisionQuery, struct ns1__ResponseType *ns1__Response) { return 0; }

